import os
import math
import json
import time
import boto3
import litellm
import logging
import pandas as pd
from datetime import datetime
from typing import Dict, Optional, List
from litellm import completion, token_counter
from fmbench.scripts.stream_responses import get_response_stream
from fmbench.scripts.fmbench_predictor import (FMBenchPredictor,
                                               FMBenchPredictionResponse)


# set a logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Initialize the global variables, including the embeddings model declaration
# and the service name
EMBEDDING_MODELS: List[str] = ["amazon.titan-embed-text-v1",
                               "cohere.embed-english-v3",
                               "cohere.embed-multilingual-v3"]
SERVICE_NAME: str = 'bedrock'


class BedrockPredictor(FMBenchPredictor):

    # initialize the service name
    _service_name: str = SERVICE_NAME

    # overriding abstract method
    def __init__(self,
                 endpoint_name: str,
                 inference_spec: Optional[Dict],
                 metadata: Optional[Dict]):
        try:
            # initialize private member variables
            self._endpoint_name = endpoint_name
            self._pt_model_id = None
            self._inference_spec = inference_spec
            self._aws_region = boto3.Session().region_name

            # check if the endpoint name corresponded to a provisioned throughput
            # endpoint
            if ':provisioned-model/' in self._endpoint_name:
                logger.info(f"{self._endpoint_name} is a provisioned throughput endpoint")
                bedrock_client = boto3.client(SERVICE_NAME)
                response = bedrock_client.list_provisioned_model_throughputs()
                if response['ResponseMetadata']['HTTPStatusCode'] != 200:
                    logger.error(f"error received while calling list_provisioned_model_throughputs, response=\"{response}\", "
                                 f"BedrockPredictor cannot be created")
                else:
                    fm_arn = [pt_summary['foundationModelArn'] for \
                                pt_summary in response['provisionedModelSummaries'] \
                                  if pt_summary['provisionedModelArn'] == self._endpoint_name]
                    if len(fm_arn) > 0:                        
                        # set the PT name which looks like arn:aws:bedrock:us-east-1:<account-id>:provisioned-model/<something>
                        self._pt_model_id = self._endpoint_name
                        # set the endpoint name which needs to look like the FM model id
                        # this can now be extracted from the fm_arn which looks like 
                        # 'arn:aws:bedrock:us-east-1::foundation-model/anthropic.claude-3-sonnet-20240229-v1:0:28k',
                        self._endpoint_name = fm_arn[0].split("/")[1]
                        logger.info(f"a matching PT was found, self._pt_model_id={self._pt_model_id}, "
                                    f"self._endpoint_name={self._endpoint_name}")
                    else:
                        logger.error(f"no matching PT was found, BedrockPredictor cannot be created")

            # model_id for the litellm API with the specific bedrock model of choice
            # endpoint_name in case of bedrock refers to the model_id such as 
            # cohere.command-text-v14 for example
            self._bedrock_model = f"{self._service_name}/{self._endpoint_name}"
            # litellm supports the following inference params as per
            # https://litellm.vercel.app/docs/completion/input
            self._temperature = 0.1
            self._max_tokens = 100
            self._top_p = 0.9
            # not used for now but kept as placeholders for future
            self._stream = None
            self._start = None
            self._stop = None

            # no caching of responses since we want every inference
            # call to be independant
            self._caching = False

            # override these defaults if there is an inference spec provided
            if inference_spec:
                parameters: Optional[Dict] = inference_spec.get('parameters')
                if parameters:
                    self._temperature = parameters.get('temperature', self._temperature)
                    self._max_tokens = parameters.get('max_tokens', self._max_tokens)
                    self._top_p = parameters.get('top_p', self._top_p)
                    self._stream = inference_spec.get("stream", self._stream)
                    self._stop = inference_spec.get("stop_token", self._stop)
                    self._start = inference_spec.get("start_token", self._start)
            self._response_json = {}
            logger.info(f"__init__, _bedrock_model={self._bedrock_model}, self._pt_model_id={self._pt_model_id},"
                        f"_temperature={self._temperature} "
                        f"_max_tokens={self._max_tokens}, _top_p={self._top_p} "
                        f"_stream={self._stream}, _stop={self._stop}, _caching={self._caching}")
        except Exception as e:
            exception_msg = f"""exception while creating predictor/initializing variables
                            for endpoint_name={self._endpoint_name}, exception=\"{e}\", 
                            BedrockPredictor cannot be created"""
            logger.error(exception_msg)
            raise ValueError(exception_msg)

    def get_prediction(self, payload: Dict) -> FMBenchPredictionResponse:
        # Represents the prompt payload
        prompt_input_data = payload['inputs']
        os.environ["AWS_REGION_NAME"] = self._aws_region
        latency: Optional[float] = None
        completion_tokens: Optional[int] = None
        prompt_tokens: Optional[int] = None
        response_dict_from_streaming: Optional[Dict] = None
        TTFT: Optional[float] = None
        TPOT: Optional[float] = None
        TTLT: Optional[float] = None

        try:
            # recording latency for when streaming is enabled
            st = time.perf_counter()
            # this response is for text generation models on bedrock: Claude, Llama, Mistral etc.
            logger.info(f"Invoking {self._bedrock_model} to get inference")
            # cohere does not support top_p and apprarently LiteLLM does not
            # know that?
            if 'cohere' not in self._endpoint_name:
                response = completion(model=self._bedrock_model,
                                      model_id=self._pt_model_id,
                                      messages=[{"content": prompt_input_data,
                                                 "role": "user"}],
                                      temperature=self._temperature,
                                      max_tokens=self._max_tokens,
                                      top_p=self._top_p,
                                      caching=self._caching,
                                      stream=self._stream)
            else:
                response = completion(model=self._bedrock_model,
                                      model_id=self._pt_model_id,
                                      messages=[{"content": prompt_input_data,
                                                 "role": "user"}],
                                      temperature=self._temperature,
                                      max_tokens=self._max_tokens,
                                      caching=self._caching,
                                      stream=self._stream)
            logger.info(f"stop token: {self._stop}, streaming: {self._stream}, "
                        f"response: {response}")

            # Get the response and the TTFT, TPOT, TTLT metrics if the streaming
            # for responses is set to true
            if self._stream is True:
                response_dict_from_streaming = get_response_stream(response,
                                                                   st,
                                                                   self._start,
                                                                   self._stop,
                                                                   is_sagemaker=False)
                TTFT = response_dict_from_streaming.get('TTFT')
                TPOT = response_dict_from_streaming.get('TPOT')
                TTLT = response_dict_from_streaming.get('TTLT')
                response = response_dict_from_streaming['response']
                self._response_json["generated_text"] = json.loads(response)[0].get('generated_text')
                # Getting in the total input and output tokens using token counter.
                # Streaming on liteLLM does not support prompt tokens and completion tokens 
                # in the invocation response format
                prompt_tokens = token_counter(model=self._endpoint_name,
                                              messages=[{"content": prompt_input_data,
                                                         "role": "user"}])
                completion_tokens = token_counter(text=self._response_json["generated_text"])
                # Extract latency in seconds
                latency = time.perf_counter() - st
                logger.info(f"streaming prompt token count: {prompt_tokens}, "
                            f"completion token count: {completion_tokens}, latency: {latency}")
            # If streaming is set to false, then get the response in the normal
            # without streaming format from LiteLLM
            else:
                # Iterate through the entire model response
                # Since we are not sending batched requests so we only expect a single completion
                for choice in response.choices:
                    # Extract the message and the message's content from LiteLLM
                    if choice.message and choice.message.content:
                        # Extract the response from the dict
                        self._response_json["generated_text"] = choice.message.content
                        break

                # Extract number of input and completion prompt tokens
                prompt_tokens = response.usage.prompt_tokens
                completion_tokens = response.usage.completion_tokens
                # Extract latency in seconds
                latency = response._response_ms / 1000
        except Exception as e:
            logger.error(f"exception during prediction, endpoint_name={self._endpoint_name}, "
                         f"exception={e}")

        return FMBenchPredictionResponse(response_json=self._response_json,
                                         latency=latency,
                                         time_to_first_token=TTFT,
                                         time_per_output_token=TPOT,
                                         time_to_last_token=TTLT,
                                         completion_tokens=completion_tokens,
                                         prompt_tokens=prompt_tokens)

    def calculate_cost(self,
                       instance_type: str,
                       instance_count: int,
                       pricing: Dict,
                       duration: float,
                       prompt_tokens: int,
                       completion_tokens: int) -> float:
        """Represents the function to calculate the cost for Bedrock experiments.
        instance_type represents the model name
        """

        # Initializing all cost variables
        experiment_cost: Optional[float] = None
        input_token_cost: Optional[float] = None
        output_token_cost: Optional[float] = None
        try:
            if self._pt_model_id is None:
                logger.info("calculate_cost, calculating cost with token based pricing")
                # Retrieve the pricing information for the instance type
                bedrock_pricing = pricing['pricing']['token_based']
                # Calculate cost based on the number of input and output tokens
                model_pricing = bedrock_pricing.get(instance_type, None)
                if model_pricing:
                    input_token_cost = (prompt_tokens / 1000.0) * model_pricing['input-per-1k-tokens']
                    output_token_cost = (completion_tokens / 1000.0) * model_pricing['output-per-1k-tokens']
                    experiment_cost = input_token_cost + output_token_cost
                    logger.info(f"instance_type={instance_type}, prompt_tokens={prompt_tokens}, "
                                f"input_token_cost={input_token_cost}, output_token_cost={completion_tokens}, "
                                f"output_token_cost={output_token_cost}, experiment_cost={experiment_cost}")
                else:
                    logger.error(f"model pricing for \"{instance_type}\" not found, "
                                 f"cannot calculate experiment cost")
            else:
                logger.info("calculate_cost, calculating cost with PT pricing")
                instance_based_pricing = pricing['pricing']['instance_based']
                hourly_rate = instance_based_pricing.get(instance_type, None)
                # calculating the experiment cost for instance based pricing
                duration_in_hours_ceil = math.ceil(duration/3600)
                experiment_cost = hourly_rate * duration_in_hours_ceil
                logger.info(f"instance_type={instance_type}, hourly_rate={hourly_rate}, "
                            f"duration_in_hours_ceil={duration_in_hours_ceil}, experiment_cost={experiment_cost}")

        except Exception as e:
            logger.error(f"exception occurred during experiment cost calculation, exception={e}")
        return experiment_cost

    def get_metrics(self,
                    start_time: datetime,
                    end_time: datetime,
                    period: int = 60) -> pd.DataFrame:
        # not implemented
        return None

    @property
    def endpoint_name(self) -> str:
        """The endpoint name property."""
        return self._endpoint_name

    @property
    def inference_parameters(self) -> Dict:
        """The inference parameters property."""
        return dict(temperature=self._temperature,
                    max_tokens=self._max_tokens,
                    top_p=self._top_p)


def create_predictor(endpoint_name: str, inference_spec: Optional[Dict], metadata: Optional[Dict]):
    if endpoint_name in EMBEDDING_MODELS:
        logger.error(f"embeddings models not supported for now")
        return None
    else:
        return BedrockPredictor(endpoint_name, inference_spec, metadata)


import os
import yaml
import boto3
import requests
import tempfile
from enum import Enum
from pathlib import Path
from datetime import datetime
from fmbench import defaults


FMBENCH_PACKAGE_NAME: str = "fmbench"

current_working_directory: str = Path.cwd()

CONFIG_FILEPATH_FILE: str = current_working_directory / 'config_filepath.txt'

# S3 client initialization
s3_client = boto3.client('s3')
session = boto3.session.Session()
region_name = session.region_name
if region_name is None:
    print(f"boto3.session.Session().region_name is {region_name}, "
          f"going to use an metadata api to determine region name")
    # THIS CODE ASSUMED WE ARE RUNNING ON EC2, for everything else
    # the boto3 session should be sufficient to retrieve region name
    resp = requests.put("http://169.254.169.254/latest/api/token",
                        headers={"X-aws-ec2-metadata-token-ttl-seconds": "21600"})
    token = resp.text
    region_name = requests.get("http://169.254.169.254/latest/meta-data/placement/region",
                               headers={"X-aws-ec2-metadata-token": token}).text
    print(f"region_name={region_name}, also setting the AWS_DEFAULT_REGION env var")
    os.environ["AWS_DEFAULT_REGION"] = region_name
print(f"region_name={region_name}")

# Configuring the role ARN -- extract the role name
caller = boto3.client('sts').get_caller_identity()
account_id = caller.get('Account')
role_arn_from_env = os.environ.get('FMBENCH_ROLE_ARN')
if role_arn_from_env:
    print(f"role_arn_from_env={role_arn_from_env}, using it to set arn_string")
    arn_string = role_arn_from_env
else:
    print(f"role_arn_from_env={role_arn_from_env}, using current sts caller identity to set arn_string")
    arn_string = caller.get('Arn')
    # if this is an assumed role then remove the assumed role related pieces
    # because we are also using this role for deploying the SageMaker endpoint
    # arn:aws:sts::015469603702:assumed-role/SSMDefaultRoleForOneClickPvreReporting/i-0c5bba16a8b3dac51
    # should be converted to arn:aws:iam::015469603702:role/SSMDefaultRoleForOneClickPvreReporting
    if ":assumed-role/" in arn_string:
        role_name = arn_string.split("/")[-2]
        arn_string = f"arn:aws:iam::{account_id}:role/{role_name}"
        print(f"the sts role is an assumed role, setting arn_string to {arn_string}")
    else:
        arn_string = caller.get('Arn')

ROLE_NAME = arn_string.split('/')[-1]

current_config_file = os.environ.get("CONFIG_FILE_FMBENCH")
# if var is true, use that from cli
if current_config_file is not None:
    CONFIG_FILE = current_config_file
    print(f"config file current -> {CONFIG_FILE}, {current_config_file}")
else:
    CONFIG_FILE = Path(CONFIG_FILEPATH_FILE).read_text().strip()
    print(f"config file current -> {CONFIG_FILE}, {current_config_file}")


if CONFIG_FILE.startswith("s3://"):
    # Parse the S3 URL
    bucket_name, key_path = CONFIG_FILE.replace("s3://", "").split("/", 1)
    # Use boto3 to access the S3 service
    s3 = boto3.resource('s3')
    # Fetch the object from S3
    obj = s3.Object(bucket_name, key_path)
    # Read the object's content
    CONFIG_FILE_CONTENT = obj.get()['Body'].read().decode('utf-8')
elif CONFIG_FILE.startswith("https://"):
    # Fetch the content from the HTTPS URL
    response = requests.get(CONFIG_FILE)
    response.raise_for_status()  # Ensure we got a successful response
    CONFIG_FILE_CONTENT = response.text
else:
    CONFIG_FILE_CONTENT = Path(CONFIG_FILE).read_text()

# check if the file is still parameterized and if so replace the parameters with actual values
# if the file is not parameterized then the following statements change nothing
write_bucket = os.environ.get("WRITE_BUCKET", f"{defaults.DEFAULT_BUCKET_WRITE}-{region_name}-{account_id}")
args = dict(region=session.region_name,
            role_arn=arn_string,
            read_tmpdir=os.path.join(tempfile.gettempdir(), defaults.DEFAULT_LOCAL_READ),
            write_tmpdir=os.path.join(tempfile.gettempdir(), defaults.DEFAULT_LOCAL_WRITE),
            write_bucket=write_bucket,
            read_bucket=f"{defaults.DEFAULT_BUCKET_READ}-{region_name}-{account_id}")
CONFIG_FILE_CONTENT = CONFIG_FILE_CONTENT.format(**args)

# Load the configuration
config = yaml.safe_load(CONFIG_FILE_CONTENT)
local_mode = os.environ.get("LOCAL_MODE")
if local_mode == "yes":
    print("globals.py, local_mode = yes")
    config['aws']['s3_and_or_local_file_system'] = 'local'
    config['s3_read_data']['s3_or_local_file_system'] = 'local'
    if config['s3_read_data'].get('local_file_system_path') is None:
        config['s3_read_data']['local_file_system_path'] = os.path.join(tempfile.gettempdir(), defaults.DEFAULT_LOCAL_READ)
    if config['aws'].get('local_file_system_path') is None:
        config['aws']['local_file_system_path'] = os.path.join(tempfile.gettempdir(), defaults.DEFAULT_LOCAL_WRITE)

# iterate through each experiment and populate the parameters section in the inference spec
for i in range(len(config['experiments'])):
    # for the experiment at index i, look up the parameter set
    # retrieve the parameter set from the inference_parameter section
    # assign the parameters from that parameter set to a new key called
    # parameters in that experiment
    parameters = config['inference_parameters'][config['experiments'][i]['inference_spec']['parameter_set']]
    config['experiments'][i]['inference_spec']['parameters'] = parameters
    if config['experiments'][i].get('bucket') is None:
        config['experiments'][i]['bucket'] = config['aws']['bucket']
print(f"loaded config: {config}")

# data directory and prompts
PER_ACCOUNT_DIR: str = f"{config['general']['name']}-{ROLE_NAME}"
DATA_DIR: str = os.path.join(PER_ACCOUNT_DIR, config['dir_paths']['data_prefix'])
PROMPTS_DIR = os.path.join(DATA_DIR, config['dir_paths']['prompts_prefix'])

# Metrics directory based on date and time
current_time = datetime.now()

# Assuming current_time is a datetime object
formatted_time = current_time.strftime("%Y/%m/%d/%H/%M")

# Split the formatted_time into components
year, month, day, hour, minute = formatted_time.split('/')

# Construct the METRICS_DIR path
METRICS_DIR = f"{DATA_DIR}/metrics/yyyy={year}/mm={month}/dd={day}/hh={hour}/mm={minute}"

METRICS_PER_INFERENCE_DIR = os.path.join(METRICS_DIR, "per_inference")
METRICS_PER_CHUNK_DIR = os.path.join(METRICS_DIR, "per_chunk")

METRICS_PER_INFERENCE_DIR = os.path.join(METRICS_DIR, "per_inference")
METRICS_PER_CHUNK_DIR = os.path.join(METRICS_DIR, "per_chunk")
ENDPOINT_METRICS_FNAME = "endpoint_metrics.csv"
ENDPOINT_METRICS_SUMMARIZED_FNAME = "endpoint_metrics_summarized.csv"

# Models directory based on date and time 
MODELS_DIR = f"{DATA_DIR}/models"

# Use this to upload to the s3 bucket (extracted from the config file)
BUCKET_NAME = config['aws']['bucket']
READ_BUCKET_NAME = config['s3_read_data']['read_bucket']

# S3 prefix
PREFIX_NAME = config['dir_paths']['data_prefix']

# SOURCE data is where your actual data resides in s3
SOURCE_DATA = config['s3_read_data']['source_data_prefix']

# Read the prompt template that the user uploads
PROMPT_TEMPLATE_S3_PREFIX = config['s3_read_data']['prompt_template_dir']

# Initialize the scripts directory
SCRIPTS_DIR: str = "fmbench/scripts"

# METADATA DIR TO HANDLE DYNAMIC S3 PATHS FOR METRICS/RESULTS
METADATA_DIR:str = config['dir_paths']['metadata_dir']
METRICS_PATH_FNAME: str = "metrics_path.txt"

DIR_LIST = [DATA_DIR, PROMPTS_DIR, METRICS_DIR, MODELS_DIR, METRICS_PER_INFERENCE_DIR, METRICS_PER_CHUNK_DIR]

# this is for custom tokenizers
TOKENIZER_DIR_S3 = config['s3_read_data']['tokenizer_prefix']
TOKENIZER = 'tokenizer'

DEPLOYMENT_SCRIPT_S3 = config['s3_read_data']['scripts_prefix']

_ = list(map(lambda x: os.makedirs(x, exist_ok=True), DIR_LIST))

# Define the endpoint list as the config-general name plus the role arn for unique generation 
# from different roles in the same/different accounts
ENDPOINT_LIST_PATH: str = os.path.join(MODELS_DIR, "endpoints.json")

REQUEST_PAYLOAD_FPATH: str = os.path.join(PROMPTS_DIR, "payload.jsonl")
RESULTS_FPATH: str = os.path.join(METRICS_DIR, "results.csv")


class TRUNCATE_POLICY(str, Enum):
    AT_PROMPT_TOKEN_LENGTH = 'at-prompt-token-length'

# misc. metrics related
PLACE_HOLDER: int = -1705338041
RESULTS_DIR: str = f"results-{config['general']['name']}"

# metric filenames
COUNTS_FNAME: str = "experiment_counts.csv"
ERROR_RATES_FNAME: str = "error_rates.csv"
RESULTS_DESC_MD_FNAME: str = "report.md"
SUMMARY_METRICS_W_PRICING_FNAME: str = "summary_metrics_w_pricing.csv"
INSTANCE_PRICING_PER_HOUR_FNAME: str = "instance_pricing_per_hour.csv"
SUMMARY_METRICS_FOR_DATASET_W_SCORES_FNAME: str = "summary_metrics_for_dataset_w_scores.csv"
SUMMARY_METRICS_FOR_DATASET_W_SCORES_BEST_OPTION_FNAME: str = "summary_metrics_for_dataset_best_option.csv"
SUMMARY_METRICS_FOR_DATASET_W_SCORES_BEST_OPTION_EACH_INSTANCE_TYPE_FNAME: str = "summary_metrics_for_dataset_best_option_each_instance_type.csv"
SUMMARY_MODEL_ENDPOINT_COST_PER_INSTANCE: str = "endpoint_per_instance_per_run_costs.csv"
BUSINESS_SUMMARY_PLOT_FNAME: str = "business_summary.png"
BUSINESS_SUMMARY_PLOT_FNAME2: str = "business_summary_barchart.png"
LATENCY_CHART_PLOT_FNAME: str = "latency_summary_chart.png"

# plot filenames
ERROR_RATES_PLOT_TEXT: str = "Error rates for different concurrency levels and instance types"
ERROR_RATES_PLOT_FNAME: str = "error_rates.png"
TOKENS_VS_LATENCY_PLOT_TEXT: str = "Tokens vs latency for different concurrency levels and instance types"
TOKENS_VS_LATENCY_PLOT_FNAME: str = "tokens_vs_latency.png"
CONCURRENCY_VS_INFERENCE_LATENCY_PLOT_FNAME: str = "concurrency_vs_inference_latency.png"
CONCURRENCY_VS_INFERENCE_LATENCY_PLOT_TEXT: str = "Concurrency Vs latency for different instance type for selected dataset"
LATENCY_BUDGET: int = 5

OVERALL_RESULTS_MD: str = """
# {title}

|**Last modified (UTC)** | **FMBench version**  |
|---|---|
|{dttm}|{fmbench_version}|


## Summary

{business_summary}

## Per instance results

The following table provides the best combinations for running inference for different sizes prompts on different instance types. The following dataset(s) were used for this test: {datasets}.

|Dataset   | Instance type   | Recommendation   |
|---|---|---|
"""

# Dataset=`{dataset}`, instance_type=`{instance_type}`
RESULT_DESC: str = """The best option for staying within a latency budget of `{latency_budget} seconds` on a `{instance_type}` for the `{dataset}` dataset is a `concurrency level of {concurrency}`. A concurrency level of {concurrency} achieves an `median latency of {latency_median} seconds`, for an `average prompt size of {prompt_size} tokens` and `completion size of {completion_size} tokens` with `{tpm} transactions/minute`."""

RESULT_ROW: str = "|`{dataset}`|`{instance_type}`|{desc}|"

RESULT_FAILURE_DESC: str = """This experiment did not find any combination of concurrency level and other configuration settings that could provide a response within a latency budget of `{latency_budget} seconds` on a `{instance_type}` for the `{dataset}` dataset."""

PROMPT_TEMPLATE: str = """<s>[INST] <<SYS>>
You are an assistant for question-answering tasks. Use the following pieces of retrieved context in the section demarcated by "```" to answer the question. If you don't know the answer just say that you don't know. Use three sentences maximum and keep the answer concise.
<</SYS>>

```
{context}
```

Question: {input}

[/INST]
Answer:

"""



import re
import os
import yaml
import math
import boto3
import logging
import requests
import tempfile
import posixpath
import unicodedata
from pathlib import Path
from fmbench import globals
from fmbench import defaults
from typing import Dict, List, Tuple
from transformers import AutoTokenizer
from botocore.exceptions import NoCredentialsError
import shutil
import concurrent.futures

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# utility functions
def load_config(config_file) -> Dict:
    """
    Load configuration from a local file or an S3 URI.

    :param config_file: Path to the local file or S3 URI (s3://bucket/key)
    :return: Dictionary with the loaded configuration
    """
    session = boto3.session.Session()
    region_name = session.region_name
    if region_name is None:
        print(f"boto3.session.Session().region_name is {region_name}, "
              f"going to use an metadata api to determine region name")
        resp = requests.put("http://169.254.169.254/latest/api/token",
                            headers={"X-aws-ec2-metadata-token-ttl-seconds": "21600"})
        token = resp.text
        region_name = requests.get("http://169.254.169.254/latest/meta-data/placement/region",
                                   headers={"X-aws-ec2-metadata-token": token}).text
        os.environ["AWS_DEFAULT_REGION"] = region_name
    print(f"region_name={region_name}")
    caller = boto3.client('sts').get_caller_identity()
    account_id = caller.get('Account')
    arn_string = caller.get('Arn')
    role_arn_from_env = os.environ.get('FMBENCH_ROLE_ARN')
    if role_arn_from_env:
        print(f"role_arn_from_env={role_arn_from_env}, using it to set arn_string")
        arn_string = role_arn_from_env
    else:
        print(f"role_arn_from_env={role_arn_from_env}, using current sts caller identity to set arn_string")
        # if this is an assumed role then remove the assumed role related pieces
        # because we are also using this role for deploying the SageMaker endpoint
        # arn:aws:sts::015469603702:assumed-role/SSMDefaultRoleForOneClickPvreReporting/i-0c5bba16a8b3dac51
        # should be converted to arn:aws:iam::015469603702:role/SSMDefaultRoleForOneClickPvreReporting
        if ":assumed-role/" in arn_string:
            role_name = arn_string.split("/")[-2]
            arn_string = f"arn:aws:iam::{account_id}:role/{role_name}"
            print(f"the sts role is an assumed role, setting arn_string to {arn_string}")
    
    # check if the file is still parameterized and if so replace the parameters with actual values
    # if the file is not parameterized then the following statements change nothing
    write_bucket = os.environ.get("WRITE_BUCKET", f"{defaults.DEFAULT_BUCKET_WRITE}-{region_name}-{account_id}")
    args = dict(region=session.region_name,
                role_arn=arn_string,
                read_tmpdir=os.path.join(tempfile.gettempdir(), defaults.DEFAULT_LOCAL_READ),
                write_tmpdir=os.path.join(tempfile.gettempdir(), defaults.DEFAULT_LOCAL_WRITE),
                write_bucket=write_bucket,
                read_bucket=f"{defaults.DEFAULT_BUCKET_READ}-{region_name}-{account_id}")

    # Check if config_file is an S3 URI
    if config_file.startswith("s3://"):
        try:
            # Parse S3 URI
            s3_client = boto3.client('s3')
            bucket, key = config_file.replace("s3://", "").split("/", 1)

            # Get object from S3 and load YAML
            response = s3_client.get_object(Bucket=bucket, Key=key)
            content = response["Body"].read().decode('utf-8')
            
        except NoCredentialsError:
            print("AWS credentials not found.")
            raise
        except Exception as e:
            print(f"Error loading config from S3: {e}")
            raise
    # Check if config_file is an HTTPS URL
    elif config_file.startswith("https://"):
        try:
            response = requests.get(config_file)
            response.raise_for_status()  # Raises a HTTPError if the response was an error
            content = response.text
        except requests.exceptions.RequestException as e:
            print(f"Error loading config from HTTPS URL: {e}")
            raise
    else:
        # Assume local file system if not S3 or HTTPS
        try:
            content = Path(config_file).read_text()
        except Exception as e:
            print(f"Error loading config from local file system: {e}")
            raise

    content = content.format(**args)
    config = yaml.safe_load(content)
    return config

def load_main_config(config_file) -> Dict:
    config = load_config(config_file)
    # iterate through each experiment and populate the parameters section in the inference spec
    for i in range(len(config['experiments'])):
        # for the experiment at index i, look up the parameter set
        # retrieve the parameter set from the inference_parameter section
        # assign the parameters from that parameter set to a new key called
        # parameters in that experiment
        parameters = config['inference_parameters'][config['experiments'][i]['inference_spec']['parameter_set']]
        config['experiments'][i]['inference_spec']['parameters'] = parameters
        if config['experiments'][i].get('bucket') is None:
            config['experiments'][i]['bucket'] = config['aws']['bucket']
    local_mode = os.environ.get("LOCAL_MODE")
    if local_mode == "yes":
        print("utils.py, local_mode = yes")
        config['aws']['s3_and_or_local_file_system'] = 'local'
        config['s3_read_data']['s3_or_local_file_system'] = 'local'
        if config['s3_read_data'].get('local_file_system_path') is None:
            config['s3_read_data']['local_file_system_path'] = os.path.join(tempfile.gettempdir(), defaults.DEFAULT_LOCAL_READ)
        if config['aws'].get('local_file_system_path') is None:
            config['aws']['local_file_system_path'] = os.path.join(tempfile.gettempdir(), defaults.DEFAULT_LOCAL_WRITE)
    return config
    
def count_tokens(text: str) -> int:
    global _tokenizer
    return _tokenizer.count_tokens(text)

def process_item(item, prompt_template_keys: List, prompt_fmt: str) -> Dict:
    args = {}
    for k in prompt_template_keys:
        v = _normalize(item[k])
        args[k] = v
        args[f"{k}_len"] = _tokenizer.count_tokens(v)
    prompt = prompt_fmt.format(**args)
    prompt_len = count_tokens(prompt)
    return args | {
        "prompt": prompt,
        "prompt_len": prompt_len
    }

def nt_to_posix(p: str) -> str:
    return p.replace("\\", "/")

def is_read_local() -> str:
    is_read_local = globals.config.get('s3_read_data').get('s3_or_local_file_system')
    logger.debug(f"is_read_local: {is_read_local}")
    return is_read_local is not None and is_read_local == 'local'


def _is_write_local_only():
    is_write_local_only = globals.config.get('aws').get('s3_and_or_local_file_system')
    logger.debug(f"is_write_local_only: {is_write_local_only}")
    return is_write_local_only is not None and is_write_local_only == 'local'

def _upload_file_to_local(local_path: str, s3_path: str) -> None:
    dest = _get_local_write_path(s3_path)
    shutil.copy(local_path, dest)

def upload_file_to_s3(bucket: str, local_path: str, s3_path: str) -> None:
    if _is_write_local_or_both():
        _upload_file_to_local(local_path, s3_path)
    if _is_write_local_only():
        return
    
    s3 = boto3.resource('s3')
    try:
        s3.Bucket(bucket).upload_file(local_path, s3_path)
    except Exception as e:
        logger.error(f"upload_file_to_s3, An error occurred: {e}")

def _write_to_local(data, dir1, dir2, file_name):
    dir = _get_local_write_path(dir1 + "/" + dir2 + "/")
    Path(dir).mkdir(parents=True, exist_ok=True)
    file = dir + file_name
    if type(data) == str:
        Path(file).write_text(data)
    else:
        Path(file).write_bytes(data)

# Function to write data to S3
def write_to_s3(data, bucket_name, dir1, dir2, file_name):
    if _is_write_local_or_both():
        _write_to_local(data, dir1, dir2, file_name)
    if _is_write_local_only():
        return

    # Initialize S3 client
    s3_client = boto3.client('s3')

    # Construct the S3 file path
    s3_file_path = posixpath.join(nt_to_posix(dir1), nt_to_posix(dir2), file_name)
    logger.debug(f"write_to_s3, s3_file_path={s3_file_path}")
    try:
        # Write the JSON data to the S3 bucket
        s3_client.put_object(Bucket=bucket_name, Key=s3_file_path, Body=data)
        return (f"s3://{bucket_name}/{s3_file_path}")
    except NoCredentialsError:
        logger.error("write_to_s3, Error: AWS credentials not found.")
    except Exception as e:
        logger.error(f"write_to_s3, An error occurred: {e}")


# Function to concurrently write multiple data to S3
# input_list: List[Tuple[data, bucket_name, dir1, dir2, file_name]
def write_multiple_to_s3(input_list: List[Tuple[None, str, str, str, str]]) -> None:
    with concurrent.futures.ThreadPoolExecutor(max_workers = 8) as executor:
        tasks = []
        for input_tuple in input_list:
            tasks.append(
                executor.submit(lambda: write_to_s3(*input_tuple))
            )

    for completed_task in concurrent.futures.as_completed(tasks):
        completed_task.result()


def _read_from_local(s3_file_path: str) -> str:
    try:
        s3_file_path = nt_to_posix(_get_local_read_path(s3_file_path))
        logger.debug(f"get_local_object, key={s3_file_path}")
        return Path(s3_file_path).read_bytes().decode('utf-8')
    except FileNotFoundError as e:
        logger.error(f"read_from_local, An error occurred: {e}")
        return None

## function to read from s3
def read_from_s3(bucket_name, s3_file_path):
    if is_read_local():
        return _read_from_local(s3_file_path)

    # Initialize S3 client
    s3_client = boto3.client('s3')
    s3_file_path = nt_to_posix(s3_file_path)

    try:
        # Fetch the object from S3
        logger.debug(f"read_from_s3, reading file from bucket={bucket_name}, key={s3_file_path}")
        response = s3_client.get_object(Bucket=bucket_name, Key=s3_file_path)
        
        return response['Body'].read().decode('utf-8')
    except NoCredentialsError:
        logger.error("read_from_s3, Error: AWS credentials not found.")
        return None
    except Exception as e:
        logger.error(f"read_from_s3, An error occurred: {e}")
        return None

def _get_local_object(bucket: str, key: str, decode: bool) -> str:
    key = nt_to_posix(key)
    logger.debug(f"get_local_object, key={key}")
    if bucket == globals.config['s3_read_data']['read_bucket']:
        pathname = _get_local_read_path(key)
    else:
        pathname = _get_local_write_path(key)
    if decode:
        return Path(pathname).read_bytes().decode('utf-8')
    else:
        return Path(pathname).read_bytes()

## gets a single s3 file
def get_s3_object(bucket: str, key: str, decode='True') -> str:
    if is_read_local():
        return _get_local_object(bucket, key, decode)

    key = nt_to_posix(key)
    logger.debug(f"get_s3_object, bucket_name={bucket}, key={key}")

    # Create an S3 client
    s3_client = boto3.client('s3')

    # Retrieve the object from S3
    response = s3_client.get_object(Bucket=bucket, Key=key)

    # Read the content of the file
    if decode:
        content = response['Body'].read().decode('utf-8')
    else:
        content = response['Body'].read()

    return content

def _list_local_files(bucket, prefix, suffix):
    if bucket == globals.config['s3_read_data']['read_bucket']:
        dir = _get_local_read_path(prefix)
    else:
        dir = _get_local_write_path(prefix)
    path_list = list(Path(dir).glob('*' + suffix))
    pathname_list = [str(item) for item in path_list]
    if bucket == globals.config['s3_read_data']['read_bucket']:
        return_list = [item.replace(_get_local_read_path(), '') for item in pathname_list]
    else:
        return_list = [item.replace(_get_local_write_path(), '') for item in pathname_list]
    return return_list

# Function to list files in S3 bucket with a specific prefix
def list_s3_files(bucket, prefix, suffix='.json'):
    if is_read_local():
        return _list_local_files(bucket, prefix, suffix)

    filter_key_by_suffix = lambda k,s: True if s is None else k.endswith(s)
    s3_client = boto3.client('s3')
    next_continuation_token = None
    
    return_list = []
    while True:
        if next_continuation_token is not None:
            response = s3_client.list_objects_v2(Bucket=bucket, Prefix=nt_to_posix(prefix), ContinuationToken=next_continuation_token)
        else:
            response = s3_client.list_objects_v2(Bucket=bucket, Prefix=nt_to_posix(prefix))
        return_list += [item['Key'] for item in response.get('Contents', []) if filter_key_by_suffix(item['Key'], suffix) is True]
        logger.info(f"found {len(return_list)} items in bucket={bucket}, prefix={prefix}, suffix={suffix}")
        if response['IsTruncated'] is True:            
            next_continuation_token = response['NextContinuationToken']
        else:
            break
    logger.info(f"there are total of {len(return_list)} items in bucket={bucket}, prefix={prefix}, suffix={suffix}")
    return return_list

def _normalize(text, form='NFC'):
    # The files in LongBench contain nonstandard or irregular Unicode.
    # For compatibility and safety we normalize them.
    return unicodedata.normalize(form, str(text))


def _is_write_local_or_both():
    is_write_local_or_both = globals.config.get('aws').get('s3_and_or_local_file_system')
    logger.debug(f"is_write_local_or_both: {is_write_local_or_both}")
    return is_write_local_or_both is not None and (is_write_local_or_both == 'local' or is_write_local_or_both == 'both')

def _get_local_read_path(dir_or_file: str = None) -> str:
    if dir_or_file is not None:
        local_read_path = globals.config['s3_read_data']['local_file_system_path'] + '/' + dir_or_file
    else:
        local_read_path = globals.config['s3_read_data']['local_file_system_path'] + '/'
    logger.debug(f"local_read_path: {local_read_path}")
    return local_read_path

def _get_local_write_path(dir_or_file: str = None) -> str:
    if dir_or_file is not None:
        local_write_path = globals.config['aws']['local_file_system_path'] + '/' + dir_or_file
    else:
        local_write_path = globals.config['aws']['local_file_system_path'] + '/'
    logger.debug(f"local_write_path: {local_write_path}")
    return local_write_path

def _download_multiple_files_from_local_write_path(prefix, local_dir):    
    src = _get_local_write_path(prefix)
    print(f"_download_multiple_files_from_local_write_path, prefix={prefix}, src={src}, local_dir={local_dir}")
    shutil.copytree(src, local_dir, dirs_exist_ok=True)

def _download_multiple_files_from_local_read_path(prefix, local_dir):    
    src = _get_local_read_path(prefix)
    print(f"_download_multiple_files_from_local_read_path, prefix={prefix}, src={src}, local_dir={local_dir}")
    shutil.copytree(src, local_dir, dirs_exist_ok=True)

def download_multiple_files_from_s3(bucket_name, prefix, local_dir):
    if _is_write_local_or_both():
        if bucket_name == globals.config['aws']['bucket']:
            return _download_multiple_files_from_local_write_path(prefix, local_dir)
        elif bucket_name == globals.config['s3_read_data']['read_bucket']:
            return _download_multiple_files_from_local_read_path(prefix, local_dir)
        else:
            logger.error(f"bucket_name={bucket_name} which does not match write bucket={globals.config['aws']['bucket']} "
                         f"or read bucket={globals.config['s3_read_data']['read_bucket']}")

    """Downloads files from an S3 bucket and a specified prefix to a local directory."""
    logger.info(f"download_multiple_files_from_s3, bucket_name={bucket_name}, prefix={prefix}, local_dir={local_dir}")
    s3_client = boto3.client('s3')

    # Ensure the local directory exists
    if not os.path.exists(local_dir):
        os.makedirs(local_dir)

    # List and download files
    try:
        response = s3_client.list_objects_v2(Bucket=bucket_name, Prefix=prefix)
        key_list = list_s3_files(bucket_name, prefix, suffix=None)
        for file_key in key_list:
            logger.debug(f"file_key={file_key}, prefix={prefix}")
            local_file_key = file_key.replace(prefix, "")
            parent_dir_in_s3 = os.path.dirname(local_file_key)
            logger.debug(f"local_file_key={local_file_key}, parent_dir_in_s3={parent_dir_in_s3}")
            # the first char for parent_dir_in_s3 would always be a '/' so skip that
            local_dir_to_create = os.path.join(local_dir, parent_dir_in_s3[1:])
            os.makedirs(local_dir_to_create, exist_ok=True)
            logger.debug(f"local_dir_to_create={local_dir_to_create}, local_file_key={local_file_key}")
            local_file_to_create = os.path.basename(local_file_key)
            if file_key.endswith('/'):
                logger.info(f"skipping file_key={file_key}")
                continue

            local_file_path = os.path.join(local_dir_to_create, local_file_to_create)
            logger.debug(f"bucket_name={bucket_name}, file_key={file_key}, local_file_path={local_file_path}")
            s3_client.download_file(bucket_name, file_key, local_file_path)
            logger.debug(f"download_multiple_files_from_s3, Downloaded: {local_file_path}")
    except Exception as e:
        logger.error(f"An error occurred while downloading from S3: {e}")

class CustomTokenizer:
    """A custom tokenizer class"""
    TOKENS: int = 1000
    WORDS: int = 750

    def __init__(self, bucket, prefix, local_dir):
        print(f"CustomTokenizer, based on HF transformers, {bucket} "
              f"prefix: {prefix} local_dir: {local_dir}")
        # Check if the tokenizer files exist in s3 and if not, use the autotokenizer
        download_multiple_files_from_s3(bucket, prefix, local_dir)
        # Load the tokenizer from the local directory
        all_files = [f for f in list(Path(local_dir).iterdir()) if f.name != ".keep"]
        dir_not_empty = len(all_files)
        print(f"CustomTokenizer, all_files = {all_files}")
        abs_path = Path(local_dir).absolute().resolve()
        if dir_not_empty > 0:
            print(f"loading the provided tokenizer from local_dir={local_dir}, abs_path={abs_path}")
            self.tokenizer = AutoTokenizer.from_pretrained(local_dir)
            print("successfully loaded the tokenizer using AutoTokenizer.from_pretrained")
        else:
            print(f"no tokenizer provided, the {local_dir}, abs_path={abs_path} is empty, "
                  f"using default tokenizer i.e. {self.WORDS} words = {self.TOKENS} tokens")
            self.tokenizer = None

    def count_tokens(self, text):
        if self.tokenizer is not None:
            return len(self.tokenizer.encode(text))
        else:
            return int(math.ceil((self.TOKENS/self.WORDS) * len(text.split())))

_tokenizer = CustomTokenizer(globals.READ_BUCKET_NAME, globals.TOKENIZER_DIR_S3, globals.TOKENIZER)



"""
Deploys a model from HuggingFace on Amazon SageMaker using the DJL DeepSpeek LMI
(https://github.com/deepjavalibrary/djl-serving).

1. Configuration is read from the configured serving.properties file.
2. A hf_token.txt file is required to download the model from Hugging Face.
"""
# Import necessary libraries
import os
import glob
import time
import boto3
import logging
import tarfile
import tempfile
import sagemaker
from pathlib import Path
from urllib.parse import urlparse
from sagemaker.utils import name_from_base
from huggingface_hub import snapshot_download
from typing import Dict, List, Tuple, Optional


# set a logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# globals
HF_TOKEN_FNAME: str = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                   "hf_token.txt")

# Initialize your S3 client for your model uploading
s3_client = boto3.client('s3')

# session/account specific variables
sess = sagemaker.session.Session()
# Define the location of your s3 prefix for model artifacts
region: str =sess._region_name

# bucket to house model artifacts
default_bucket = sess.default_bucket()

# Define your account id
account_id: str = sess.account_id()

# Initialize the sagemaker and sagemaker runtime clients
sm_client = boto3.client("sagemaker")

def _download_model(model_id: str,
                    local_model_path: str,
                    allow_patterns: Optional[List] = ["*"]) -> str:
    """
    Download the model files locally
    """
    local_model_path = Path(local_model_path)
    print(f"Local model path: {local_model_path}")
    local_model_path.mkdir(exist_ok=True)
    print(f"Created the local directory: {local_model_path}")

    model_download_path = snapshot_download(
        repo_id=model_id,
        cache_dir=local_model_path,
        allow_patterns=allow_patterns,
        use_auth_token=Path(HF_TOKEN_FNAME).read_text().strip()
    )
    print(f"Uncompressed model downloaded into ... -> {model_download_path}")
    return model_download_path

def _upload_dir(localDir: str, awsInitDir: str, bucketName: str, tag: str ="*.*"):
    s3 = boto3.resource('s3')
    p = Path(localDir)
    # Iterate over all directories and files within localDir
    for path in p.glob('**/*'):
        if path.is_file():
            rel_path = path.relative_to(p)
            awsPath = os.path.join(awsInitDir, str(rel_path)).replace("\\", "/")
            logger.info(f"Uploading {path} to s3://{bucketName}/{awsPath}")
            logger.info(f"path: {path}, bucket name: {bucketName}, awsPath: {awsPath}")
            s3.meta.client.upload_file(path, bucketName, awsPath)

def _create_and_upload_model_artifact(serving_properties_path: str,
                                      bucket: str,
                                      prefix: str) -> str:
    """
    Create the model artifact with the updated serving properties within the directory
    """
    # Create a tar.gz file containing only the serving.properties file
    tar_file_path = os.path.join(Path(serving_properties_path).parent, 'model.tar.gz')
    with tarfile.open(tar_file_path, "w:gz") as tar:
        # Add the serving.properties file
        tar.add(serving_properties_path, arcname='serving.properties')

    # Upload the tar.gz file to S3
    key = f"{prefix}/model.tar.gz"
    s3_client.upload_file(tar_file_path, bucket, key)
    model_tar_gz_path: str = f"s3://{bucket}/{key}"
    logger.info(f"uploaded model.tar.gz to {model_tar_gz_path}")
    return model_tar_gz_path

def _create_model(experiment_config: Dict,
                  inference_image_uri: str,
                  s3_model_artifact: str,
                  role_arn: str) -> Tuple[str, str]:
    """
    # Function to create the SageMaker model
    """
    model_name = name_from_base(experiment_config['model_name'])
    env = experiment_config.get('env')

    # HF token required for gated model downloads form HF
    hf_token_file_path = Path(HF_TOKEN_FNAME)
    hf_dict: Optional[Dict] = None
    if hf_token_file_path.is_file() is True:
        logger.info(f"hf_token file path: {hf_token_file_path} is a file")
        hf_dict = dict(HUGGING_FACE_HUB_TOKEN=hf_token_file_path.read_text().strip())
    else:
        logger.info(f"hf_token file path: {hf_token_file_path} is not a file")
        
    # this gets passed as an env var
    if env:
        if hf_dict:
            # both env and hf_dict exists, so we do a union
            env = env | hf_dict
    else:
        if hf_dict:
            # env var did not exist, but hf_dict did so that
            # is now the env var
            env = hf_dict

    if env:
        pc = dict(Image=inference_image_uri,
                  ModelDataUrl=s3_model_artifact,
                  Environment=env)
    else:
        pc = dict(Image=inference_image_uri,
                  ModelDataUrl=s3_model_artifact)

    create_model_response = sm_client.create_model(
        ModelName=model_name,
        ExecutionRoleArn=role_arn,
        PrimaryContainer=pc,
    )
    return model_name, create_model_response["ModelArn"]


def _deploy_endpoint(experiment_config: Dict,
                     model_name: str) -> Tuple[str, str]:
    """
    Function to create and deploy the endpoint
    """
    endpoint_config_name = f"{model_name}-config"
    endpoint_name = f"{model_name}-endpoint"

    _ = sm_client.create_endpoint_config(
        EndpointConfigName=endpoint_config_name,
        ProductionVariants=[
            {
                "VariantName": "variant1",
                "ModelName": model_name,
                "InstanceType": experiment_config["instance_type"],
                "InitialInstanceCount": 1,
                "ModelDataDownloadTimeoutInSeconds": 3600,
                "ContainerStartupHealthCheckTimeoutInSeconds": 3600,
            },
        ],
    )

    create_endpoint_response = sm_client.create_endpoint(
        EndpointName=endpoint_name, EndpointConfigName=endpoint_config_name
    )
    return endpoint_name, create_endpoint_response['EndpointArn']


def _check_endpoint_status(endpoint_name: str) -> str:
    """
    Function to check the status of the endpoint
    """
    resp = sm_client.describe_endpoint(EndpointName=endpoint_name)
    status = resp["EndpointStatus"]
    while status == "Creating":
        time.sleep(60)
        resp = sm_client.describe_endpoint(EndpointName=endpoint_name)
        status = resp["EndpointStatus"]
    return status


def deploy(experiment_config: Dict, role_arn: str) -> Dict:
    """
    Function to deploy the model and create the endpoint
    """

    if experiment_config.get("download_from_hf_place_in_s3") is True:
        with tempfile.TemporaryDirectory() as local_model_path:
            logger.info(f"created temporary directory {local_model_path}")
        local_model_path = _download_model(experiment_config['model_id'],
                                           local_model_path)
        logger.info(f"going to upload model files to {experiment_config['model_s3_path']}")

        o = urlparse(experiment_config['model_s3_path'], allow_fragments=False)
        _upload_dir(local_model_path, o.path.lstrip('/'), o.netloc) 
        logger.info(f"local model path: {local_model_path}, o.path: {o.path}, o.netloc: {o.netloc}")

        model_artifact = experiment_config['model_s3_path']
        logger.info(f"Uncompressed model downloaded into ... -> {model_artifact}")

    logger.info("preparing model artifact...")

    # handle serving.properties, we read it from the config and then write it to
    # a local file
    write_bucket = experiment_config.get('bucket', default_bucket)
    logger.info(f"write bucket for inserting model.tar.gz into: {write_bucket}")
    properties = experiment_config["serving.properties"].format(write_bucket=write_bucket)
    dir_path = os.path.dirname(os.path.realpath(__file__))
    serving_properties_path = os.path.join(dir_path, "serving.properties")
    Path(serving_properties_path).write_text(properties)
    logger.info(f"written the following serving.properties \
                  content={properties} to {serving_properties_path}")

    # create and upload the model.tar.gz, note that this file is just a placeholder
    # it is not the actual model, the actual model binaries are in s3 or HuggingFace
    # and the container will download them when the model endpoint is being created
    logger.info(f"uploading model.tar.gz to S3,bucket={write_bucket}, \
                  prefix={experiment_config['model_id']}")
    model_artifact = _create_and_upload_model_artifact(serving_properties_path,
                                                      write_bucket,
                                                      experiment_config['model_id'])
    logger.info(f"model uploaded to: {model_artifact}")

    inference_image_uri = experiment_config['image_uri']
    logger.info(f"inference image URI: {inference_image_uri}")

    # create model
    model_name, model_arn = _create_model(experiment_config,
                                          inference_image_uri,
                                          model_artifact,
                                          role_arn)
    logger.info(f"created Model: {model_arn}")

    # deploy model
    endpoint_name, _ = _deploy_endpoint(experiment_config, model_name)
    logger.info(f"deploying endpoint: {endpoint_name}")

    # check model deployment status
    status = _check_endpoint_status(endpoint_name)
    logger.info(f"Endpoint status: {status}")

    if status == 'InService':
        logger.info("endpoint is in service")
    else:
        logger.info("endpoint is not in service.")

    return dict(endpoint_name=endpoint_name,
                experiment_name=experiment_config['name'],
                instance_type=experiment_config['instance_type'],
                instance_count=experiment_config['instance_count'],
                deployed=True)



import os
import re
import sys
import yaml
import json
import time
import boto3
import logging
import argparse
import requests
import papermill as pm
from typing import Dict
from pathlib import Path
from datetime import datetime
from nbformat import NotebookNode

# Setup logging
logging.basicConfig(format='[%(asctime)s] p%(process)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s', level=logging.INFO)
logger = logging.getLogger(__name__)

# Create an S3 client
s3_client = boto3.client('s3')

# Function to check whether the given uri is a valid s3 uri
def is_valid_s3_uri(s3_uri: str) -> bool:
    pattern = re.compile(r's3://[^/]+/.+')
    return bool(pattern.match(s3_uri))

# Function to check whether the given uri is a valid https URL
def is_valid_http_url(url: str) -> bool:
    return url.startswith("https://") or url.startswith("http://")

# Function to get the s3_uri from the user and get the config file path, writing the txt file
def read_config(config_file_path: str) -> Dict:
    if is_valid_s3_uri(config_file_path):
        logger.info(f"executing the config file found in {config_file_path}...")

        bucket, key = config_file_path.replace("s3://", "").split("/", 1)

        # Get object from S3 and load YAML
        response = s3_client.get_object(Bucket=bucket, Key=key)
        config_content = yaml.safe_load(response["Body"])
    elif is_valid_http_url(config_file_path):
        try:
            logger.info(f"loading config from HTTPS URL: {config_file_path}")
            response = requests.get(config_file_path)
            response.raise_for_status()  # Raises a HTTPError for bad responses
            config_content = yaml.safe_load(response.text)
        except requests.exceptions.RequestException as e:
            logger.error(f"error loading config from HTTPS URL: {e}")
            raise
    else:
        logger.info(f"the provided URI '{config_file_path}' is not a valid S3 URI or HTTPS URL, assuming this is a local file")
        config_content = yaml.safe_load(Path(config_file_path).read_text())

    # You can choose to write to config_path here if needed, otherwise just return the loaded content
    logger.info(f"loaded configuration: {json.dumps(config_content, indent=2)}")
    return config_content

# Function to handle cell outputs
def output_handler(cell: NotebookNode, _):
    if cell.cell_type == 'code':
        for output in cell.get('outputs', []):
            if output.output_type == 'stream':
                print(output.text, end='')


def run_notebooks(config_file: str) -> None:
    # Assume `read_config` function is defined elsewhere to load the config
    config = read_config(config_file)

    current_directory = Path(__file__).parent
    logging.info(f"Current directory is --> {current_directory}")

    output_directory = current_directory / "executed_notebooks"
    if not output_directory.exists():
        output_directory.mkdir()

    for step, execute in config['run_steps'].items():
        if execute:
            notebook_path = current_directory / step
            logging.info(f"Current step file --> {notebook_path.stem}")

            timestamp = datetime.now().strftime("%Y%m%d-%H%M%S")
            output_file = output_directory / f"{notebook_path.stem}_{timestamp}.ipynb"

            try:
                logging.info(f"Executing {notebook_path.name}...")
                logger.info(f"THE STEP BEING EXECUTED NOW: {step}")
                pm.execute_notebook(
                    input_path=str(notebook_path),
                    output_path=str(output_file),
                    kernel_name='python3',
                    parameters={},
                    report_mode=True,  
                    progress_bar=True,
                    stdout_file=None, 
                    stderr_file=None,
                    log_output=True,
                    output_handler=output_handler 
                )
                logger.info(f"STEP EXECUTION COMPLETED: {step}")
            except FileNotFoundError as e:
                logging.error(f"File not found: {e.filename}")
                sys.exit(1)
            except Exception as e:
                logging.error(f"Failed to execute {step}: {str(e)}")
                sys.exit(1)
        else:
            logging.info(f"Skipping {step} as it is not marked for execution")

    logger.info(f"FMBench has completed the benchmarking process. Check S3 bucket \"{config['aws']['bucket']}\" for results")


# main function to run all of the fmbench process through a single command via this python package
def main():
    parser = argparse.ArgumentParser(description='Run FMBench with a specified config file.')
    parser.add_argument('--config-file', type=str, help='The S3 URI of your Config File', required=True)
    role_help = 'The ARN of the role to be used for FMBench. If an \
                 Amazon SageMaker endpoint is being deployed \
                 through FMBench then this role would also be used \
                 by that endpoint'
    parser.add_argument('--role-arn', type=str, default=None, required=False, help=role_help)
    # add an option to run FMBench local mode. If local mode is set to yes, then FMBench uses read and write data locally and if no, 
    # then the test will continue to interact with S3
    parser.add_argument('--local-mode', type=str, default=None, choices=['yes', 'no'], help='Specify if running in local mode or not. Options: yes, no. Default is no.')
    parser.add_argument('--write-bucket', type=str, help='Write bucket that is used for sagemaker endpoints in local mode and storing metrics in s3 mode.')

    args = parser.parse_args()
    print(f"main, {args} = args")

    # Set the environment variable based on the parsed argument
    os.environ["CONFIG_FILE_FMBENCH"] = args.config_file
    logger.info(f"Config file specified: {args.config_file}")
    
    # set env var to indicate that fmbench is being run from main and not interactively via a notebook
    os.environ["INTERACTIVE_MODE_SET"] = "no"

    # set the environment variable for the local mode option
    if args.local_mode:
        print(f"setting the LOCAL_MODE to {args.local_mode}")
        os.environ["LOCAL_MODE"] = args.local_mode
        if args.local_mode == "yes":
            if args.write_bucket is None:
                logger.error("Write bucket is not provided when local mode is set to 'yes'")
                sys.exit(1)
            else:
                # set the environment variable for the write bucket name to be configured and used for sagemaker endpoints and 
                # metrics stored in s3 mode with local mode is set to "yes" by the user
                os.environ["WRITE_BUCKET"] = args.write_bucket
                logger.info(f"Write bucket specified in local mode: {args.write_bucket}")



    # if a role arn is specified then set it as an env var so that the rest of the code
    # can use it. This will then be used to set the templatized "sagemaker_execution_role"
    # parameter in the config file
    if args.role_arn:
        print(f"setting FMBENCH_ROLE_ARN env variable to {args.role_arn}")
        os.environ['FMBENCH_ROLE_ARN'] = args.role_arn

    # Proceed with the rest of your script's logic, passing the config file as needed
    run_notebooks(args.config_file)    

if __name__ == "__main__":
    main()




"""
Deploys a model from HuggingFace on Amazon EC2

1. Configuration is read from the configured serving.properties file.
2. A hf_token.txt file is required to download the model from Hugging Face.
"""
# Import necessary libraries
import os
import sys
import time
import json
import stat
import docker
import logging
import requests
import tempfile
import subprocess
from typing import Dict
from pathlib import Path

# set a logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# globals
HF_TOKEN_FNAME: str = os.path.join(os.path.dirname(os.path.realpath(__file__)), "hf_token.txt")
# SHM_SIZE: str = "12g"
# MODEL_DEPLOYMENT_TIMEOUT: int = 2400
FMBENCH_MODEL_CONTAINER_NAME: str = "fmbench_model_container"

def _set_up(model_name: str, serving_properties: str, local_model_path: str):
    """
    Create the model serving.properties file locally in a model directory
    """
    # make the model directory with serving.properties
    directory = os.path.join(local_model_path, model_name)
    os.makedirs(directory, exist_ok=True)
    logger.info(f"Creating the local directory: {directory}")

    file_path = os.path.join(directory, "serving.properties")
    Path(file_path).write_text(serving_properties)
    logger.info(f"The serving.properties file has been created in {file_path}")
    st = os.stat(file_path)
    os.chmod(file_path, st.st_mode | stat.S_IEXEC)
    # Make the serving.properties file executable
    logger.info(f"chmod happening")
    logger.info(f"The serving.properties file has been made executable.")

    #return the directory we created
    return directory

def _create_deployment_script(image_uri, region, model_name, HF_TOKEN, directory, gpu_or_neuron_setting, model_loading_timeout):
    """
    Write a deployment script for model container
    """
    #stop container if it already exists check if container exists 
    container_name: str = FMBENCH_MODEL_CONTAINER_NAME
    deploy_script_content = f"""#!/bin/sh
echo "Going to download model now"
echo "Content in docker command: {region}, {image_uri}, {model_name}"
aws ecr get-login-password --region {region} | docker login --username AWS --password-stdin {image_uri}
docker pull {image_uri}
docker stop {container_name} || true && docker rm {container_name} || true
docker run -d --name={container_name} \\
 {gpu_or_neuron_setting} \\
 -v {directory}:/opt/ml/model:ro \\
 -v {directory}/model_server_logs:/opt/djl/logs \\
 -e MODEL_LOADING_TIMEOUT={model_loading_timeout} \\
 -e HF_TOKEN={HF_TOKEN} \\
 -p 8080:8080 \\
 {image_uri}
echo "Done pulling model"
"""
    script_file_path = os.path.join(directory, "deploy_model.sh")
    Path(script_file_path).write_text(deploy_script_content)
    logger.info(f"deploy_model.sh content: {deploy_script_content}")
    logger.info(f"The 'deploy_model.sh' file has been created in {script_file_path}")
    return script_file_path

def _run_container(script_file_path):
    """
    Runs the deploy_model.sh bash script with the provided arguments.
    """
    logger.info(f"Running container script at {script_file_path}")
    # Create a Docker client
    client = docker.from_env()

    try:
        # Check if the container exists and is running
        container = client.containers.get(FMBENCH_MODEL_CONTAINER_NAME)
        if container.status == "running":
            logger.info(f"Container {FMBENCH_MODEL_CONTAINER_NAME} is already running.")
        else:
            logger.info(f"Container {FMBENCH_MODEL_CONTAINER_NAME} is not running. Running the script directly.")
            subprocess.run(["bash", script_file_path], check=True)
            logger.info(f"done running bash script")
        return True
    except docker.errors.NotFound:
        logger.info(f"Container {FMBENCH_MODEL_CONTAINER_NAME} not found. Running the script directly.")
        subprocess.run(["bash", script_file_path], check=True)
        logger.info(f"done running bash script")
        return True
    except subprocess.CalledProcessError as e:
        logger.error(f"Error running deploy_model.sh script: {e}")
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")
    return False

def _check_model_deployment(endpoint, model_loading_timeout):
    """
    Check the model deployment status and wait for the model to be ready.
    """
    start_time = time.time()
    #global variable 
    timeout = model_loading_timeout
    logger.info(f"Checking deployment status at {endpoint}")
    data = {"inputs": ["tell me a story of the little red riding hood"]}
    headers = {"content-type": "application/json"}
    while time.time() - start_time < timeout:
        time.sleep(60)
        try:
            response = requests.post(endpoint, headers=headers, json=data)
            logger.info(f"response is: {response.text}")
            if response.status_code == 200:
                logger.info("Model deployment successful!")
                return "InService"
            else:
                print(f"Model deployment is not ready yet. Return code: {response.status_code}")
                time.sleep(60)  # Wait for 1 minute before checking again
        except Exception as e:
            logger.error(f"Error occurred while deploying the endpoint: {e}")
            time.sleep(60)  # Wait for 1 minute before checking again
    return "Failed"

def deploy(experiment_config: Dict, role_arn: str) -> Dict:
    """
    Function to deploy the model and create the endpoint
    """
    image_uri: str = experiment_config['image_uri']
    model_name: str = experiment_config['name']
    logger.info(f"Going to deploy model: {model_name}")
    ep_name: str = experiment_config['ep_name']
    model_id: str = experiment_config['model_id']
    region: str = experiment_config['region']
    serving_properties: str = experiment_config['serving.properties']
    model_loading_timeout: int = experiment_config['ec2']['model_loading_timeout']
    gpu_or_neuron_setting: str = experiment_config['ec2']['gpu_or_neuron_setting']
    HF_TOKEN: str = Path(HF_TOKEN_FNAME).read_text().strip()
    dir_path = home_dir = os.getenv("HOME", str(Path.home()))
    logger.info(f"Home directory: {dir_path}")
    model_directory = _set_up(model_name, serving_properties, dir_path)
    logger.info(f"Writing serving.properties {serving_properties} to {model_directory}")

    logger.info("Creating the deployment script in model directory")
    deployment_script_path = _create_deployment_script(image_uri, region, model_name, HF_TOKEN, model_directory, gpu_or_neuron_setting, model_loading_timeout )

    logger.info("Running the deployment script")
    ran_container = _run_container(deployment_script_path)

    # initialize with None values for error case
    deployment_result: Dict = dict(endpoint_name=None, 
                        experiment_name=None,
                        instance_type=None,
                        instance_count=None, 
                        deployed=False)
    if ran_container:
        logger.info("Container ran successfully")
        ep_status = _check_model_deployment(ep_name, model_loading_timeout)
        logger.info(f"Endpoint status: {ep_status}")
        if ep_status == "InService":
            logger.info("Model endpoint running!")
            deployment_result['endpoint_name'] = ep_name
            deployment_result['experiment_name'] = experiment_config['name']
            deployment_result['instance_type'] = experiment_config['instance_type']
            deployment_result['instance_count'] = experiment_config['instance_count']
            deployment_result['deployed'] = True
            return deployment_result
        elif ep_status == "Failed":
            logger.error("Model endpoint not running!")
            return deployment_result
    else:
        logger.error("Container did not run successfully")
        return deployment_result



import time
import json
import boto3
import logging
import sagemaker
import pandas as pd
from datetime import datetime
from typing import Dict, Optional
from fmbench.utils import count_tokens
from sagemaker.predictor import Predictor
from sagemaker.serializers import JSONSerializer
from fmbench.scripts.stream_responses import get_response_stream
from fmbench.scripts.sagemaker_metrics import get_endpoint_metrics
from fmbench.scripts.fmbench_predictor import (FMBenchPredictor,
                                               FMBenchPredictionResponse)

# set a logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Initialize the SageMaker runtime to support getting response streams
sagemaker_runtime = boto3.client('sagemaker-runtime')


class SageMakerPredictor(FMBenchPredictor):
    # overriding abstract method
    def __init__(self,
                 endpoint_name: str,
                 inference_spec: Optional[Dict],
                 metadata: Optional[Dict]):
        self._predictor: Optional[sagemaker.base_predictor.Predictor] = None
        self._endpoint_name: str = endpoint_name
        self._inference_spec: Dict = inference_spec
        self._variant_name: Optional[str] = None
        self._use_messages_api_format: Optional[bool] = None
        if metadata is not None:
            self._variant_name = metadata.get("variant_name")
            self._use_messages_api_format = metadata.get("use_messages_api_format")

        try:
            # Create a SageMaker Predictor object
            self._predictor = Predictor(
                endpoint_name=self._endpoint_name,
                sagemaker_session=sagemaker.Session(),
                serializer=JSONSerializer()
            )
        except Exception as e:
            logger.error(f"create_predictor, exception occured while creating predictor "
                         f"for endpoint_name={self._endpoint_name}, exception={e}")
        logger.info(f"__init__ _predictor={self._predictor}, "
                    f"_variant_name={self._variant_name}_"
                    f"inference_spec={self._inference_spec}, "
                    f"_use_messages_api_format={self._use_messages_api_format}")

    def get_prediction(self, payload: Dict) -> FMBenchPredictionResponse:
        response_json: Optional[Dict] = None
        response: Optional[str] = None
        latency: Optional[float] = None
        TTFT: Optional[float] = None
        TPOT: Optional[float] = None
        TTLT: Optional[float] = None
        prompt_tokens: Optional[int] = None
        completion_tokens: Optional[int] = None
        streaming: Optional[bool] = None
        stop_token: Optional[str] = None

        # represents the number of tokens in the prompt payload
        prompt_tokens = count_tokens(payload["inputs"])

        try:
            st = time.perf_counter()
            split_input_and_inference_params = None
            if self._inference_spec is not None:
                split_input_and_inference_params = self._inference_spec.get("split_input_and_parameters")
            response = None
            streaming = self._inference_spec.get("stream", False)
            if split_input_and_inference_params is True:
                response = self._predictor.predict(payload["inputs"],
                                                   self._inference_spec["parameters"])
            else:
                if self._use_messages_api_format is True:
                    # if needed in future add support for system prompt as well
                    # and multiple system/user prompts but all that is not needed for now
                    payload = {"messages": [{"role": "user",
                                             "content": payload["inputs"]}]}
                    # the final payload should look like this:
                    # {
                    #   "top_p": 0.9,
                    #   "max_tokens": 100,
                    #   "messages": [
                    #     {
                    #       "role": "user",
                    #       "content": "this is the prompt"
                    #     }
                    #   ]
                    # }
                    payload = payload | dict(self._inference_spec["parameters"])
                else:
                    # the final payload should look like this:
                    # {
                    #   "parameters": { 
                    #     "top_p": 0.9,
                    #     "max_tokens": 100
                    #    },
                    #   "inputs": "this is the prompt"
                    # }
                    payload = payload | dict(parameters=self._inference_spec["parameters"])

            # if the response streaming is step, call the get_response stream on the 
            # sagemaker endpoint, else use the simple predict call
            if streaming is True:
                start_token = self._inference_spec.get("start_token", None)
                stop_token = self._inference_spec.get("stop_token", None)
                payload["stream"] = streaming
                logger.info(f"streaming={streaming}, calling invoke_endpoint_with_response_stream")
                response_stream = sagemaker_runtime.invoke_endpoint_with_response_stream(
                                                    EndpointName=self._endpoint_name,
                                                    Body=json.dumps(payload),
                                                    ContentType="application/json")
                response_dict = get_response_stream(response_stream['Body'],
                                                    st,
                                                    start_token,
                                                    stop_token,
                                                    is_sagemaker=True)
                TTFT = response_dict.get('TTFT')
                TPOT = response_dict.get('TPOT')
                TTLT = response_dict.get('TTLT')
                response = response_dict.get('response')
            else:
                logger.info(f"streaming={streaming}, calling predict")
                response = self._predictor.predict(payload)

            latency = time.perf_counter() - st
            if isinstance(response, bytes):
                response = response.decode('utf-8')
            response_json = json.loads(response)

            if isinstance(response_json, list):
                response_json = response_json[0]
            # add a key called completion, if not there
            if response_json.get("generated_text") is None:
                if response_json.get("predicted_label") is not None:
                    response_json["generated_text"] = response_json.get("predicted_label")
            # counts the completion tokens for the model using the default/user provided tokenizer
            completion_tokens = count_tokens(response_json.get("generated_text"))

        except Exception as e:
            logger.error(f"get_prediction, exception occurred while getting prediction for payload={payload} "
                         f"from predictor={self._endpoint_name}, response={response}, exception={e}")
        return FMBenchPredictionResponse(response_json=response_json,
                                         latency=latency,
                                         time_to_first_token=TTFT,
                                         time_per_output_token=TPOT,
                                         time_to_last_token=TTLT,
                                         completion_tokens=completion_tokens,
                                         prompt_tokens=prompt_tokens)

    @property
    def endpoint_name(self) -> str:
        """The endpoint name property."""
        return self._endpoint_name

    def calculate_cost(self,
                       instance_type: str,
                       instance_count: int,
                       pricing: Dict,
                       duration: float,
                       prompt_tokens: int,
                       completion_tokens: int) -> float:
        """Calculate the cost of each experiment run."""
        experiment_cost: Optional[float] = None
        try:
            instance_based_pricing = pricing['pricing']['instance_based']
            hourly_rate = instance_based_pricing.get(instance_type, None)
            logger.info(f"the hourly rate for running on {instance_type} is {hourly_rate}, "
                        f"instance_count={instance_count}")
            # calculating the experiment cost for instance based pricing
            instance_count = instance_count if instance_count else 1
            experiment_cost = (hourly_rate / 3600) * duration * instance_count
        except Exception as e:
            logger.error(f"exception occurred during experiment cost calculation, exception={e}")
        return experiment_cost

    def get_metrics(self,
                    start_time: datetime,
                    end_time: datetime,
                    period: int = 60) -> pd.DataFrame:
        return get_endpoint_metrics(self._endpoint_name, self._variant_name, start_time, end_time)
        
    @property
    def inference_parameters(self) -> Dict:
        """The inference parameters property."""
        return self._inference_spec.get("parameters")


def create_predictor(endpoint_name: str, inference_spec: Optional[Dict], metadata: Optional[Dict]):
    return SageMakerPredictor(endpoint_name, inference_spec, metadata)



# import the required libraries
import os
import time
import json
import logging
import sagemaker
import subprocess
from pathlib import Path
from typing import Dict, Optional

# session/account specific variables
sess = sagemaker.session.Session()
# Define the location of your s3 prefix for model artifacts
region: str =sess._region_name
HF_TOKEN_FNAME: str = os.path.join(os.path.dirname(os.path.realpath(__file__)), "hf_token.txt")

# set a logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def _init_eks_checks(eks_cluster_name: str):
    """
    This function describes the EKS cluster, and updates the kubeconfig for the cluser 
    before deploying the model
    """
    try:
        # Describe the EKS Cluster
        logger.info("Describing EKS Cluster...")
        describe_command_args = ["aws", "eks", "--region", region, "describe-cluster", "--name", eks_cluster_name]
        describe_result = subprocess.run(describe_command_args, capture_output=True, text=True)

        # Check if the cluster exists in the user account
        if describe_result.returncode != 0:
            logger.error("Error: EKS cluster does not exists. Please run the Terraform script")
            return
        else:
            logger.info("Describe cluster step done, cluster exists!")

        # Update the kubeconfig before deploying the model
        logger.info("Updating the kubeconfig...")
        update_command_args = ["aws", "eks", "--region", region, "update-kubeconfig", "--name", eks_cluster_name]
        update_result = subprocess.run(update_command_args, capture_output=True, text=True)
        logger.info(update_result.stdout)
        # check if the kubeconfig has been updated
        if update_result.returncode != 0:
            logger.error("Error: kubeconfg not updated")
            return
        else:
            logger.info("kubeconfig updated")

        # Check for the nodes available in the cluster
        logger.info("Showing kubectl")
        show_nodes_command_args = ["kubectl", "version"]
        show_nodes_result = subprocess.run(show_nodes_command_args, capture_output=True, text=True)
        if show_nodes_result.returncode != 0:
            logger.error("Error: nodes not shown")
            return
        else:
            logger.info(f"available kubectl version: {show_nodes_result.stdout}")

        # show the nodes
        logger.info("Showing nodes...")
        show_nodes_command_args = ["kubectl", "get", "nodes"]
        show_nodes_result = subprocess.run(show_nodes_command_args, capture_output=True, text=True)
        if show_nodes_result.returncode != 0:
            logger.error("Error: nodes not shown")
            return
        else:
            logger.info(f"Nodes shown: {show_nodes_result.stdout}")
    except Exception as e:
        logger.error(f"Error occurred while updating the kubeconfig: {e}")


def _deploy_ray(manifest_file_name: str, manifest_dir_path: str):
    """
    This function deploys the model using ray with a kubectl apply command
    """
    try:
        # check the path to the ray file within the configs/eks_manifests directory
        # manifest_ray_fpath: str = os.path.join(MANIFESTS_FOLDER, manifest_file_name)
        current_dir: str = os.path.dirname(os.path.realpath(__file__))
        parent_dir: str = os.path.abspath(os.path.join(current_dir, os.pardir))
        manifest_dir_path = os.path.join(parent_dir, manifest_dir_path)
        # Get the absolute path of the manifest file
        manifest_ray_fpath = os.path.join(manifest_dir_path, manifest_file_name)
        logger.info(f"Manifest file absolute path: {manifest_ray_fpath}")
        # HF token required for gated model downloads form HF
        hf_token_file_path = Path(HF_TOKEN_FNAME)
        if hf_token_file_path.is_file() is True:
            logger.info(f"hf_token file path: {hf_token_file_path} is a file")
            hf_token = Path(HF_TOKEN_FNAME).read_text().strip()
        else:
            logger.info(f"hf_token file path: {hf_token_file_path} is not a file")
        os.environ["AWS_REGION"] = region
        os.environ["HUGGING_FACE_HUB_TOKEN"] = hf_token
        # make sure the file is at the path
        if not os.path.isfile(manifest_ray_fpath):
            logger.error("Error: Ray file not found")
            subprocess.run("pwd")
            return
        logger.info(f"Deploying the model using the {manifest_file_name} manifest file")
        deploy_result = subprocess.run(f"envsubst < {manifest_ray_fpath} | kubectl apply -f -", shell=True, capture_output=True)
        logger.info(deploy_result.stdout)
        if deploy_result.returncode != 0:
            logger.error(f"Error: Ray not deployed: {deploy_result.stderr}")
        else:
            logger.info(f"Ray Serve Initiated: {deploy_result.stdout}")
    except Exception as e:
        logger.error(f"Error occurred while deploying the model: {e}")


def _check_ray_service_status(eks_model_namespace: str):
    """
    After the deployment step begins, this function checks the status of the model deployment
    every 15 seconds for 30 minutes. If the deployment fails, it errors out.
    """
    try:
        # Set time limit
        start_time: float = time.time()
        # 30 minutes
        timeout: int = (30 * 60)

        while (time.time() - start_time) < timeout:
            logger.info("Checking if Ray service is deployed...")
            # Check the status of the services
            check_command_args = ["kubectl", "-n", eks_model_namespace, "get", "services", "-o", "json"]
            check_result = subprocess.run(check_command_args, capture_output=True, text=True)
            # get the svc count
            status_json = json.loads(check_result.stdout)
            status_list = status_json['items']
            svc_count = len(status_list)
            # Check if all services are in the "Active" state
            if (svc_count == 3):
                logger.info("Ray service is deployed")
                break
            else:
                logger.info("Ray service is not fully deployed yet, waiting for 15 seconds...")
                logger.info(f"Current # of svc deployed: {svc_count}")
                time.sleep(15)
        else:
            logger.error("Error: Ray service is not deployed within 20 minutes.")
    except Exception as e:
        logger.error(f"Error occurred while checking the deployment status of the model: {e}")


def _print_ingress(eks_model_namespace: str, inference_url_format: str) -> str:
    """
    This function prints the endpoint url string that is returned once the model gets
    deployed
    """
    endpoint_url: Optional[str] = None
    try:
        get_command_args = ["kubectl", "-n", eks_model_namespace, "get", "ingress"]
        get_endpoint_result = subprocess.run(get_command_args, capture_output=True, text=True)
        if get_endpoint_result.returncode != 0:
            logger.error("Error: Ingress not printed")

        else:
            logger.info("Ingress Info:")
        # get the ELB ID
        get_elb_id_command_args = ["kubectl", "get", "-n", eks_model_namespace, "ingress", "-o", "jsonpath='{.items[*].status.loadBalancer.ingress[0].hostname}'"]
        get_elb_id_result = subprocess.run(get_elb_id_command_args, capture_output=True, text=True)
        elb_id: str = get_elb_id_result.stdout
        ep_api_url: str = 'http://' + elb_id[1:-1]
        # append the inference format to the endpoint url to run inferences against it
        endpoint_url = ep_api_url + inference_url_format
        logger.info(f"Inference URL: {endpoint_url}")
    except Exception as e:
        logger.error(f"Error occurred while returning the endpoint url: {e}")
        endpoint_url = None
    return endpoint_url


def deploy(experiment_config: Dict, role_arn: str) -> Dict:
    """
    This is the deploy function to run all of the deployment steps to deploy the 
    model on EKS and return a dictionary containing the endpoint url, experiment name, 
    instance type and instance count
    """
    eks_endpoint_info: Optional[Dict] = None
    try:
        # first update the kubeconfig
        eks_cluster_name: str = experiment_config['eks']['eks_cluster_name']
        _init_eks_checks(eks_cluster_name)

        # deploy the model
        manifest_file_name: str = experiment_config['eks']['manifest_file']
        manifest_dir_path: str = experiment_config['eks']['manifest_dir']
        _deploy_ray(manifest_file_name, manifest_dir_path)

        # check the status every 15 seconds during deployment
        eks_model_namespace: str = experiment_config['eks']['eks_model_namespace']
        _check_ray_service_status(eks_model_namespace)

        # fetch the endpoint url once the model is deployed
        inference_url_format: str = experiment_config['inference_spec']['inference_url_format']
        endpoint_url: str = _print_ingress(eks_model_namespace, inference_url_format)
        logger.info(f"Deployed endpoint URL: {endpoint_url}")
        eks_endpoint_info = dict(endpoint_name= endpoint_url,
                                 experiment_name=experiment_config['name'],
                                 instance_type=experiment_config['instance_type'],
                                 instance_count=experiment_config['instance_count'], 
                                 deployed=True)
    except Exception as e:
        logger.error(f"Error occured while deploying the model: {e}")
        eks_endpoint_info = None
    return eks_endpoint_info







"""
Retrieves metrics for SageMaker Endpoints from CloudWatch.
See https://docs.aws.amazon.com/sagemaker/latest/dg/monitoring-cloudwatch.html for
full list of metrics.
"""
import json
import boto3
import logging
import pandas as pd
from datetime import datetime, timedelta

# Setup logging
logging.basicConfig(format='[%(asctime)s] p%(process)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s', level=logging.INFO)
logger = logging.getLogger(__name__)

def _get_endpoint_utilization_metrics(endpoint_name: str,
                                      variant_name: str,
                                      start_time: datetime,
                                      end_time: datetime,
                                      period : int = 60) -> pd.DataFrame:
    """
    Retrieves utilization metrics for a specified SageMaker endpoint within a given time range.

    Parameters:
    - endpoint_name (str): The name of the SageMaker endpoint.
    - start_time (datetime): The start time for the metrics data.
    - end_time (datetime): The end time for the metrics data.
    - period (int): The granularity, in seconds, of the returned data points. Default is 60 seconds.

    Returns:
    - Dataframe: A Dataframe containing metric values for utilization metrics like CPU and GPU Usage.
    """
    
    metrics = ["CPUUtilization",
               "MemoryUtilization",
               "DiskUtilization",
               "InferenceLatency",
               "GPUUtilization",
               "GPUMemoryUtilization"]
    
    client = boto3.client('cloudwatch')
    data = []
    namespace = "/aws/sagemaker/Endpoints"
    
    for metric_name in metrics:
        logger.debug(f"_get_endpoint_utilization_metrics, endpoint_name={endpoint_name}, variant_name={variant_name}, "
                     f"metric_name={metric_name}, start_time={start_time}, end_time={end_time}")
        response = client.get_metric_statistics(
            Namespace=namespace,
            MetricName=metric_name,
            Dimensions=[
                {
                    'Name': 'EndpointName',
                    'Value': endpoint_name
                },
                {
                    'Name': 'VariantName',
                    'Value': variant_name
                }
            ],
            StartTime=start_time,
            EndTime=end_time,
            Period=period,
            Statistics=['Average']  # You can also use 'Sum', 'Minimum', 'Maximum', 'SampleCount'
        )
        logger.debug(response)
        for datapoint in response['Datapoints']:
            data.append({
                'EndpointName': endpoint_name, 
                'Timestamp': datapoint['Timestamp'],
                'MetricName': metric_name,
                'Average': datapoint['Average']
            })

    # Create a DataFrame from the collected data
    df = pd.DataFrame(data)

    # Pivot the DataFrame to have metrics as columns
    df_pivot = df.pivot_table(index=['Timestamp', 'EndpointName'], columns='MetricName', values='Average').reset_index()
    
    # Remove the index column heading
    sm_utilization_metrics_df = df_pivot.rename_axis(None, axis=1)
    
    return sm_utilization_metrics_df


def _get_endpoint_invocation_metrics(endpoint_name: str,
                                     variant_name: str,
                                     start_time: datetime,
                                     end_time: datetime,
                                     period : int = 60):
    """
    Retrieves Invocation metrics for a specified SageMaker endpoint within a given time range.

    Parameters:
    - endpoint_name (str): The name of the SageMaker endpoint.
    - start_time (datetime): The start time for the metrics data.
    - end_time (datetime): The end time for the metrics data.
    - period (int): The granularity, in seconds, of the returned data points. Default is 60 seconds.

    Returns:
    - Dataframe: A Dataframe containing metric values for Invocation metrics like Invocations and Model Latency.
    """
    metric_names = ["Invocations",
                    "Invocation4XXErrors",
                    "Invocation5XXErrors",
                    "ModelLatency",
                    "InvocationsPerInstance"]
    
    # Initialize a session using Amazon CloudWatch
    client = boto3.client('cloudwatch')

    namespace = "AWS/SageMaker"
    data = []
    
    for metric_name in metric_names:
        if metric_name == 'ModelLatency':
            stat = 'Average'
        else:
            stat = 'Sum'
        logger.debug(f"_get_endpoint_invocation_metrics, endpoint_name={endpoint_name}, variant_name={variant_name}, "
                     f"metric_name={metric_name}, start_time={start_time}, end_time={end_time}")
        # Get metric data for the specified metric
        response = client.get_metric_data(
            MetricDataQueries=[
                {
                    'Id': f'metric_{metric_name}',
                    'MetricStat': {
                        'Metric': {
                            'Namespace': namespace,
                            'MetricName': metric_name,
                            'Dimensions': [
                                {
                                    'Name': 'EndpointName',
                                    'Value': endpoint_name
                                },
                                {
                                    'Name': 'VariantName',
                                    'Value': variant_name
                                }
                            ]
                        },
                        'Period': period,  # Period in seconds
                        'Stat': stat  # Statistic to retrieve
                    },
                    'ReturnData': True,
                },
            ],
            StartTime=start_time,
            EndTime=end_time
        )
        logger.debug(response)
        # Extract the data points from the response
        timestamps = response['MetricDataResults'][0]['Timestamps']
        values = response['MetricDataResults'][0]['Values']
        
        for timestamp, value in zip(timestamps, values):
            data.append({
                'EndpointName': endpoint_name, 
                'Timestamp': timestamp,
                'MetricName': metric_name,
                'Value': value
            })

    # Create a DataFrame from the collected data
    df = pd.DataFrame(data)
    
    # Pivot the DataFrame to have metrics as columns
    df_pivot = df.pivot_table(index=['Timestamp', 'EndpointName'], columns='MetricName', values='Value').reset_index()
    
    # Remove the index column heading
    sm_invocation_metrics_df = df_pivot.rename_axis(None, axis=1)
    
    return sm_invocation_metrics_df


def get_endpoint_metrics(endpoint_name: str,
                         variant_name: str,
                         start_time: datetime,
                         end_time: datetime,
                         period: int = 60):
    """
    Retrieves Invocation and Utilization metrics for a specified SageMaker endpoint within a given time range.

    Parameters:
    - endpoint_name (str): The name of the SageMaker endpoint.
    - start_time (datetime): The start time for the metrics data.
    - end_time (datetime): The end time for the metrics data.
    - period (int): The granularity, in seconds, of the returned data points. Default is 60 seconds.

    Returns:
    - Dataframe: A Dataframe containing metric values for Utilization and Invocation metrics.
    """
    
    endpoint_metrics_df: Optional[pd.DataFrame] = None
    try:
        logger.info(f"get_endpoint_metrics, going to retrieve endpoint utlization metrics for "
                    f"endpoint={endpoint_name}, variant_name={variant_name}, start_time={start_time}, "
                    f"end_time={end_time}, period={period}")
        utilization_metrics_df = _get_endpoint_utilization_metrics(endpoint_name=endpoint_name,
                                                                   variant_name=variant_name,
                                                                   start_time=start_time,
                                                                   end_time=end_time,
                                                                   period=period)
        logger.info(f"get_endpoint_metrics, going to retrieve endpoint invocation metrics for "
                    f"endpoint={endpoint_name}, variant_name={variant_name}, start_time={start_time}, "
                    f"end_time={end_time}, period={period}")
        invocation_metrics_df = _get_endpoint_invocation_metrics(endpoint_name=endpoint_name,
                                                                 variant_name=variant_name,
                                                                 start_time=start_time,
                                                                 end_time=end_time,
                                                                 period=period)

        endpoint_metrics_df = pd.merge(utilization_metrics_df,
                                       invocation_metrics_df,
                                       on=['Timestamp', 'EndpointName'],
                                       how='outer')
        logger.info(f"get_endpoint_metrics, shape of invocation and utilization metrics for "
                    f"endpoint={endpoint_name} is {endpoint_metrics_df.shape}")
        logger.info(f"get_endpoint_metrics, endpoint_metrics_df={endpoint_metrics_df.head()}")
    except Exception as e:
        logger.error(f"get_endpoint_metrics, exception occured while retrieving metrics for {endpoint_name}, "
                     f"exception={e}")

    return endpoint_metrics_df


import io
import time
import json
import boto3
import litellm
import logging
import botocore
from typing import Dict, Optional, List, Union

# Set up logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Initialize the SageMaker runtime to invoke the endpoint
sagemaker_runtime = boto3.client('sagemaker-runtime')


class LineIterator:

    """
    A helper class for parsing the byte stream input.

    The output of the model will be in the following format:
    ```
    b'{"outputs": [" a"]}\n'
    b'{"outputs": [" challenging"]}\n'
    b'{"outputs": [" problem"]}\n'
    ...
    ```

    While usually each PayloadPart event from the event stream will contain a byte array 
    with a full json, this is not guaranteed and some of the json objects may be split across
    PayloadPart events. For example:
    ```
    {'PayloadPart': {'Bytes': b'{"outputs": '}}
    {'PayloadPart': {'Bytes': b'[" problem"]}\n'}}
    ```

    This class accounts for this by concatenating bytes written via the 'write' function
    and then exposing a method which will return lines (ending with a '\n' character) within
    the buffer via the 'scan_lines' function. It maintains the position of the last read 
    position to ensure that previous bytes are not exposed again. 
    """

    def __init__(self, stream):
        self.byte_iterator = iter(stream)
        self.buffer = io.BytesIO()
        self.read_pos = 0

    def __iter__(self):
        return self

    def __next__(self):
        while True:
            self.buffer.seek(self.read_pos)
            line = self.buffer.readline()
            if line and line[-1] == ord('\n'):
                self.read_pos += len(line)
                return line[:-1]
            try:
                chunk = next(self.byte_iterator)
            except StopIteration:
                if self.read_pos < self.buffer.getbuffer().nbytes:
                    continue
                raise
            if 'PayloadPart' not in chunk:
                logger.warning('Unknown event type:' + str(chunk))
                continue
            self.buffer.seek(0, io.SEEK_END)
            self.buffer.write(chunk['PayloadPart']['Bytes'])


def get_response_stream(response_stream: Union[litellm.utils.CustomStreamWrapper, botocore.eventstream.EventStream],
                        start_time: float,
                        start_token: str,
                        stop_token: str,
                        is_sagemaker: bool = False) -> Dict:
    """
    Helper function to get the response streams from bedrock or sagemaker invocations
    and parse each as appropriate to calculate the Time To First Token (TTFT), Time Per Output Token (TPOT), 
    and Time To Last Token (TTLT)

    return: This function returns a dictionary containing the entire response, and the TTFT, TTLT, TPOT metrics
    """
    logger.info(f"get_response_stream, type(response_stream)={type(response_stream)}")
    sm_start_token = b"{"
    first_token_time: Optional[float] = None
    token_times: List[float] = []
    last_token_time = start_time
    response_text: str = ""
    TTFT: Optional[float] = None
    TPOT: Optional[float] = None
    TTLT: Optional[float] = None
    result: Optional[Dict] = None

    try:
        # get the event from the sagemaker or bedrock response streams
        event_iterator = LineIterator(response_stream) if is_sagemaker else response_stream

        for event in event_iterator:
            # if the response stream is from a sagemaker call, then use the 
            # line iterator to get the first token from the streaming response
            token_id: Optional[int] = None
            if is_sagemaker:
                if event != b'' and sm_start_token in event:
                    data = json.loads(event[event.find(sm_start_token):].decode('utf-8'))
                    #logger.info(f"data={data}")
                    token_id = data['token']['id']
                    if token_id != [-1]:
                        token_text = data['token']['text']
                    else:
                        token_text = None
                else:
                    continue
            else:
                # if the response stream is from a bedrock call, then get the chunks from the response
                # and the first token
                if hasattr(event, 'choices') and hasattr(event.choices[0], 'delta'):
                    token_text = event.choices[0].delta.get('content', '')
                else:
                    continue
            # record the current time
            current_time = time.perf_counter()
            if token_text and token_text != stop_token:
                if first_token_time is None:
                    first_token_time = current_time
                    # get the time to first token latency
                    TTFT = first_token_time - start_time
                    logger.info(f"Time to First Token: {TTFT:.6f} seconds")
                else:
                    # if the token is not the first token, then get the inter-token
                    # latency
                    token_time = current_time - last_token_time
                    # append all token times to get the time per output token
                    token_times.append(token_time)
                last_token_time = current_time
                response_text += token_text

            if stop_token and stop_token in response_text:
                logger.info(f"got the last token: {stop_token}")
                break
            elif token_id == [-1]:
                logger.info(f"end of stream because token id is {token_id}")
                break

        # Calculate TTLT at the reception of the last token
        current_time = time.perf_counter()
        TTLT = current_time - start_time
        logger.info(f"Time to Last Token (TTLT): {TTLT:.6f} seconds, total tokens received {len(token_times)}")

        if token_times:
            # get the average of all token times to compute the time per output token
            TPOT = sum(token_times) / len(token_times)
            logger.info(f"Time Per Output Token (TPOT): {TPOT:.6f} seconds")

        response_data = [{"generated_text": response_text}]
        response_json_str = json.dumps(response_data)
        result = {
            "TTFT": TTFT,
            "TPOT": TPOT,
            "TTLT": TTLT,
            "response": response_json_str
        }
    except Exception as e:
        logger.error(f"Error occurred while generating and computing metrics "
                     f"associated with the streaming response: {e}", exc_info=True)
        result = None

    logger.info(f"Final result: {result}")
    return result




import os
import json
import math
import time
import boto3
import logging
import requests
import pandas as pd
from datetime import datetime
from fmbench.utils import count_tokens
from typing import Dict, Optional, List
from fmbench.scripts.fmbench_predictor import (FMBenchPredictor,
                                               FMBenchPredictionResponse)
                                            
# set a logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class EC2Predictor(FMBenchPredictor):
    # overriding abstract method
    def __init__(self,
                 endpoint_name: str,
                 inference_spec: Optional[Dict], 
                 metadata: Optional[Dict]):
        try:
            self._endpoint_name: str = endpoint_name
            self._inference_spec: Dict = inference_spec
        except Exception as e:
            logger.error(f"create_predictor, exception occured while creating predictor "
                         f"for endpoint_name={self._endpoint_name}, exception={e}")
        logger.info(f"_endpoint_name={self._endpoint_name}, _inference_spec={self._inference_spec}")

    def get_prediction(self, payload: Dict) -> FMBenchPredictionResponse:
        response_json: Optional[Dict] = None
        response: Optional[str] = None
        latency: Optional[float] = None
        TTFT: Optional[float] = None
        TPOT: Optional[float] = None
        TTLT: Optional[float] = None
        prompt_tokens: Optional[int] = None
        completion_tokens: Optional[int] = None
        # get the prompt for the EKS endpoint
        prompt: str = payload['inputs']
        # represents the number of tokens in the prompt payload
        prompt_tokens = count_tokens(payload['inputs'])
        try:
            st = time.perf_counter()
            split_input_and_inference_params: Optional[bool] = None
            if self._inference_spec is not None:
                split_input_and_inference_params = self._inference_spec.get("split_input_and_parameters")
                logger.info(f"split input parameters is: {split_input_and_inference_params}")
            # this is the POST request to the endpoint url for invocations that 
            # is given to you as you deploy a model on EC2 using the DJL serving stack
            response = requests.post(self._endpoint_name, payload) 
            # record the latency for the response generated
            latency = time.perf_counter() - st
            # For other response types, change the logic below and add the response in the `generated_text` key within the response_json dict
            response.raise_for_status()
            full_output = response.text
            answer_only = full_output.replace(prompt, "", 1).strip('["]?\n')
            response_json = dict(generated_text=answer_only)
            # counts the completion tokens for the model using the default/user provided tokenizer
            completion_tokens = count_tokens(response_json.get("generated_text"))
        except requests.exceptions.RequestException as e:
            logger.error(f"get_prediction, exception occurred while getting prediction for payload={payload} "
                         f"from predictor={self._endpoint_name}, response={response}, exception={e}")
        return FMBenchPredictionResponse(response_json=response_json,
                                         latency=latency,
                                         time_to_first_token=TTFT,
                                         time_per_output_token=TPOT,
                                         time_to_last_token=TTLT,
                                         completion_tokens=completion_tokens,
                                         prompt_tokens=prompt_tokens)

    @property
    def endpoint_name(self) -> str:
        """The endpoint name property."""
        return self._endpoint_name

    # The rest ep is deployed on an instance that incurs hourly cost hence, the calculcate cost function
    # computes the cost of the experiment on an hourly basis. If your instance has a different pricing structure
    # modify this function.
    def calculate_cost(self,
                       instance_type: str,
                       instance_count: int,
                       pricing: Dict,
                       duration: float,
                       prompt_tokens: int,
                       completion_tokens: int) -> float:
        """Calculate the cost of each experiment run."""
        experiment_cost: Optional[float] = None
        try:
            instance_based_pricing = pricing['pricing']['instance_based']
            hourly_rate = instance_based_pricing.get(instance_type, None)
            logger.info(f"the hourly rate for running on {instance_type} is {hourly_rate}, instance_count={instance_count}")
            # calculating the experiment cost for instance based pricing
            instance_count = instance_count if instance_count else 1
            experiment_cost = (hourly_rate / 3600) * duration * instance_count
        except Exception as e:
            logger.error(f"exception occurred during experiment cost calculation, exception={e}")
        return experiment_cost
    
    def get_metrics(self,
                    start_time: datetime,
                    end_time: datetime,
                    period: int = 60) -> pd.DataFrame:
        # not implemented
        return None

    @property
    def inference_parameters(self) -> Dict:
        """The inference parameters property."""
        return self._inference_spec.get("parameters")

def create_predictor(endpoint_name: str, inference_spec: Optional[Dict], metadata: Optional[Dict]):
    return EC2Predictor(endpoint_name, inference_spec, metadata)



import boto3
import logging
import sagemaker
from typing import Dict
from sagemaker import Model
from sagemaker.utils import name_from_base

logging.basicConfig(format='[%(asctime)s] p%(process)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s', level=logging.INFO)
logger = logging.getLogger(__name__)


def deploy(experiment_config: Dict, role_arn: str) -> Dict:
    role = role_arn
    endpoint_name = name_from_base(experiment_config['ep_name'])

    boto3_session=boto3.session.Session()
    smr = boto3.client('sagemaker-runtime')
    sm = boto3.client('sagemaker')
    # sagemaker session for interacting with different AWS APIs
    sess = sagemaker.session.Session(boto3_session, 
                                     sagemaker_client=sm, 
                                     sagemaker_runtime_client=smr)
    logger.info("======================================")
    logger.info(f"Will load artifacts from {experiment_config['model_data']}")
    logger.info("======================================")

    logger.info("======================================")
    logger.info(f"Using Container image {experiment_config['image_uri']}")
    logger.info("======================================")
    model = Model(name=endpoint_name,
                  # Enable SageMaker uncompressed model artifacts
                  model_data=experiment_config['model_data'],
                  image_uri=experiment_config['image_uri'],
                  role=role,
                  env=experiment_config['env'],
                  sagemaker_session=sess)
    logger.info(model)

    logger.info(f'\nModel deployment initiated Endpoint Name: {endpoint_name}\n')
    if experiment_config.get('accept_eula') is not None:
        model.deploy(
            initial_instance_count=experiment_config['instance_count'],
            instance_type=experiment_config['instance_type'],
            endpoint_name=endpoint_name,
            #volume_size=512, # not allowed for the selected Instance type ml.g5.12xlarge
            model_data_download_timeout=1200, # increase the timeout to download large model
            container_startup_health_check_timeout=1200, # increase the timeout to load large model,
            wait=True,
            accept_eula=experiment_config.get('accept_eula')
        )
    else:
        model.deploy(
            initial_instance_count=experiment_config['instance_count'],
            instance_type=experiment_config['instance_type'],
            endpoint_name=endpoint_name,
            #volume_size=512, # not allowed for the selected Instance type ml.g5.12xlarge
            model_data_download_timeout=1200, # increase the timeout to download large model
            container_startup_health_check_timeout=1200, # increase the timeout to load large model,
            wait=True
        )
    logger.info(f'Model deployment on Endpoint Name: {endpoint_name} finished\n')
    return dict(endpoint_name=endpoint_name,
                experiment_name=experiment_config['name'],
                instance_type=experiment_config['instance_type'],
                instance_count=experiment_config['instance_count'], 
                deployed=True)